/*

*/
import scene.*;
import scene.models.*;
import pipeline.*;
import framebuffer.*;

import java.io.File;
import java.awt.Color;
import java.awt.event.*;

/**
   Put the camera at the origin of world coordinates looking
   down the negative z-axis. This way, world coordinates and
   view coordinates coincide.

   Put a sphere of radius 1/4 at center (0, 0, -5/4) and put
   another sphere of radius 1/4 at center (0, 0, -7/4) so
   the second sphere is right behind and just touching the
   first sphere. Initialize the near plane with near = 1, so
   the near plane is just touching the front of the first
   sphere (recall that the near plane is z = -near in view
   coordinates).

   The n-key will move the near plane towards the first
   sphere (and the z-key will move the first sphere towards
   the near plane). Each tap of the n-key moves the near
   plane by 0.01, so it takes 50 taps to move the near plane
   through the whole first sphere. Notice that at the 25'th
   tap of the n-key, the first sphere starts to shrink in
   radius because at that point the near plane has passed
   through more than half of the sphere

   After 50 taps of the n-key, only the second sphere remains
   and the near plane is just touching it. so 50 more taps of
   the n-key will make the near plane pass through it also.

   At any time you can switch from perspective projection to
   orthographic projection and it should not change how much
   of the sphere has been clipped off.
*/
public class TestNearClippingPlane
{
   private static double near   =  1.0;
   private static double left   = -1.0;
   private static double right  =  1.0;
   private static double bottom = -1.0;
   private static double top    =  1.0;

   private static double xTranslation = 0.0;
   private static double yTranslation = 0.0;
   private static double zTranslation = 0.0;
   private static double xRotation = 0.0;
   private static double yRotation = 0.0;
   private static double zRotation = 0.0;

   private static boolean useFOVY = false;
   private static double fovy = 90.0;

   public static void main(String[] args)
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the model along the x, y, z axes.");
      System.out.println("Use the u/U, v/V, w/W, keys to rotate the model around the x, y, z axes.");
      System.out.println("Use the 'c' key to change the solid model colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the 'n' key to move the camera's near plane.");
      System.out.println("Use the 'f' key to toggle fixing the perspective field of view.");

      // Create two sphere Models with radius 1/4.
      Model m1 = new SphereModel(10, 10, 0.25);
      Model m2 = new SphereModel(50, 50, 0.25);


      // Create the Scene object that we shall render
      final Scene scene = new Scene();

      // Add models to the Scene.
      scene.addModel(m1, m2);

      // Put the sphere m1 at the edge of the initial near
      // plane (z = -1 in view coordinates).
      m1.modelTranslate(0, 0, -1.25);
      xTranslation =  0.0;
      yTranslation =  0.0;
      zTranslation = -1.25;

      // Put the sphere m2 right behind m1
      m2.modelTranslate(0, 0, -1.75);

      // Give the models random colors.
      for (Model m : scene.modelList)
         m.setColorRandom();


      // Create a Camera object.
      final Camera camera = new Camera();

      // Set up the camera's view volume.
      camera.projPerspective(left, right, bottom, top, near);
      // An alternative way to set up the camera's view frustum.
    //camera.projPerspective(90, 1, near);

      // Set up the camera's location and orientation.
      double centerX = 0, centerY = 0, centerZ = -2;
      double     upX = 0,     upY = 1,     upZ = 0;
      double    eyeX = 0,    eyeY = 0,    eyeZ = 0;
      camera.viewLookAt(eyeX,    eyeY,    eyeZ,
                        centerX, centerY, centerZ,
                        upX,     upY,     upZ);

      // Add the Camera to the Scene.
      scene.setCamera(camera);


      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      final InteractiveFrame app = new InteractiveFrame("Interactive JFrame", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );
            char c = e.getKeyChar();
            if ('d' == c)
            {
             //Pipeline.debug = ! Pipeline.debug;
               Pipeline.debug_visible = ! Pipeline.debug_visible;
             //Rasterize.debug = ! Rasterize.debug;
             //RasterizeAntialiasing.debug = ! RasterizeAntialiasing.debug;
            }
            else if ('f' == c)
            {
               useFOVY = ! useFOVY;
               if ( useFOVY )
               {
                  fovy = 2.0 * (180./Math.PI) * Math.atan(top/near);
                  System.out.println("The field of view is fixed at " + fovy + ".");
               }
               else
               {
                  top    =  near * Math.tan((Math.PI/180.0)*fovy/2.0);
                  bottom = -top;
                  right  =  top;
                  left   = -right;
                  System.out.println("The field of view will change with near plane.");
               }
            }
            else if ('n' == c || 'N' == c)
            {
               // Move the near plane.
               if ('n' == c)
                  near += 0.01;
               else
                  near -= 0.01;
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('a' == c)
            {
               Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
            }
            else if ('c' == c)
            {
               // Change the solid random color of each model.
               for (Model m : scene.modelList)
                     m.setColorRandom();
            }
            else if ('x' == c)
            {
               xTranslation += 0.1;
            }
            else if ('X' == c)
            {
               xTranslation -= 0.1;
            }
            else if ('y' == c)
            {
               yTranslation += 0.1;
            }
            else if ('Y' == c)
            {
               yTranslation -= 0.1;
            }
            else if ('z' == c)
            {
               zTranslation += 0.01;
            }
            else if ('Z' == c)
            {
               zTranslation -= 0.01;
            }
            else if ('u' == c)
            {
               xRotation += 2.0;
            }
            else if ('U' == c)
            {
               xRotation -= 2.0;
            }
            else if ('v' == c)
            {
               yRotation += 2.0;
            }
            else if ('V' == c)
            {
               yRotation -= 2.0;
            }
            else if ('w' == c)
            {
               zRotation += 2.0;
            }
            else if ('W' == c)
            {
               zRotation -= 2.0;
            }

            // Set the model-to-world transformation matrix.
            // The order of the transformations is very important!
            Model model = scene.modelList.get(0);
            model.model2Identity();
            model.modelTranslate(xTranslation, yTranslation, zTranslation);
            model.modelRotate(xRotation, 1, 0, 0);
            model.modelRotate(yRotation, 0, 1, 0);
            model.modelRotate(zRotation, 0, 0, 1);

            // Set up the camera's view volume.
            if (scene.camera.perspective)
            {
               if ( ! useFOVY )
               {
                  camera.projPerspective(left, right, bottom, top, near);
               }
               else
               {
                  // An alternative way to set up the camera's view frustum.
                  camera.projPerspective(fovy, 1, near);
               }
            }
            else
            {
               camera.projOrtho(-1, 1, -1, 1, near);
            }
            if ('n' == c || 'N' == c) System.out.print( scene.camera );

            // Render again.
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );
            java.awt.Container c = (java.awt.Container)e.getSource();
            java.awt.Insets insets = c.getInsets();
            int w = c.getWidth()  - insets.left - insets.right;
            int h = c.getHeight() - insets.top  - insets.bottom;

            this.createFrameBuffer(w, h);
            this.getFrameBuffer().clearFB(Color.black);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }
      };

      app.addKeyListener(app);
      app.addComponentListener(app);
   }
}
