/*

*/
package pipeline;
import scene.*;

/**
   Clip any (projected and normalized) line segment that sticks out of the
   standardized clipping region (including a front clipping plane) in clip
   coordinates.

   This clipping algorithm is a simplification of the Liang-Barsky Parametric
   Line Clipping algorithm.

   This version of the algorithm assumes that all LineSegment objects have
   been projected and normalized in (homogeneous) clip coordinates. This
   algorithm will clip those line segments with respect to the standardized
   clipping region (including a front clipping plane) with
      -1 <= x/w <= +1,
      -1 <= y/w <= +1, and
            z/w <= +1.

   This algorithm uses homogeneous clip coordinates, where the homogeneous
   coordinate w need NOT be 1. In order to avoid doing a lot of unneeded
   divisions, we will write the inequalities for the standard clipping
   region as
      -w <= x <= w,
      -w <= y <= w, and
            z <= w.
   It is important to realize that these homogeneous clipping inequalities
   are correct for both the perspectively projected view volume and for the
   othographically projected view volume. (These two projected view volumes
   do not equal each other in homogeneous coordinates, but they will be equal
   to each other after the perspective division.)

   If a line segment's projected and normalized vertex has an x or y coordinate
   with absolute value greater than w, then that vertex "sticks out" of the
   standard (normalized) clipping region. This algorithm will clip the
   line segment so that both of the line segment's vertices are within the
   standard clipping region with -w <= x <= w, -w <= y <= w, and z <= w..

   NOTE: Rewriting the inequality -1 <= x/w <= +1 as -w <= x <= w only makes
   sense if we know that w is a positive number. The perspective projection
   matrix takes points (x,y,z,w) in view coordinates and transforms them as
   (x,y,z,w)->(x,y,z+1,-z). But in view coordinates, z is always negative (why?),
   so -z is positive, so the w-coordinates after the perspective projection
   are always positive. (And the orthographic projection matrix transforms
   view coordinates as (x,y,z,w)->(x,y,z-1,1), so w is always positive.)


   Here is an outline of the clipping algorithm.

   Recursively process each line segment, using the following steps.

     1) Test if the line segment no longer needs to be clipped, i.e., both of
        its vertices are within the normalized clipping region. If this is
        the case, then return true.

     2) Test if the line segment should be "trivially rejected". A line segment
        is "trivially rejected" if it is on the wrong side of any of the five
        planes that bound the normalized clipping region (i.e., the five planes
        z = w, x = w, x = -w, y = w, y = -w). If this is the case, then return
        false (so the line segment will not be rasterized into the framebuffer).

        Notice that a line like the following one is trivially rejected because it
        is on the "wrong" side of the plane x = w.

                           x=w
                            |                      v1
                            |                    /
                 +----------+                  /
                 |          |                /
                 |          |              /
                 |          |            /
                 |          |          /
                 |          |        /
                 +----------+      /
                            |    /
                            |  v0

        But the following line is NOT trivially rejected because, even though it
        is completely outside of the normalized clipping region, this line is not
        entirely on the wrong side of any one of the four planes x = w, x = -w,
        y = w, or y = -w. The line below will get clipped at least one time
        (either on the plane x = w or the plane y = -w) before it is (recursively)
        a candidate for "trivial rejection". Notice that the line below could even
        be clipped twice, first on y = w, then on x = w, before it can be trivially
        rejected (by being on the wrong side of y = -w).

                           x=w
                            |                      v1
                            |                    /
                 +----------+                  /
                 |          |                /
                 |          |              /
                 |          |            /
                 |          |          /
                 |          |        /
                 +----------+      /
                            |    /
                            |  /
                             /
                           /
                        v0  |
                            |

     3) If the line segment has been neither accepted nor rejected, then it needs
        to be clipped. So we test the line segment against each of the five clipping
        planes, z = w, x = w, x = -w, y = w, and y = -w, to determine if the line segment
        crosses one of those planes. We clip the line segment against the first plane
        which we find that it crosses. Then we recursively clip the resulting clipped
        line segment. (Notice that we only clip against the first clipping plane which
        the segment is found to cross. We do not continue to test against the other
        clipping planes. This is because it may be the case, after just one clip, that
        the line segment is now a candidate for trivial accept or reject. So rather
        than test the line segment against several more clipping planes (which may be
        useless tests) it is more efficient to recursively clip the line segment, which
        will then start with the trivial accept or reject tests.)

        When we clip a line segment against a clipping plane, it is always the case that
        one endpoint of the line segment is on the "right" side of the clipping plane and
        the other endpoint of the line segment is on the "wrong" side of the clipping
        plane. In the following picture, assume that v0 is on the "wrong" side of the
        clipping plane (x = w) and v1 is on the "right" side. So v0 needs to be clipped
        off the line segment and replaced by a new endpoint.

                             x=w
                              |
                        v1    |
                          \   |
                            \ |
                              \
                              | \
                              |   \
                              |     v0

        Represent points p(t) on the line segment between v0 and v1 with the
        following parametric equation.
                  p(t) = (1-t) * v0 + t * v1  with  0 <= t <= 1
        Notice that this equation parameterizes the line segment starting with v0
        at t=0 (on the "wrong side") and ending with v1 at t=1 (on the "right side").
        We need to find the value of t when the line segment crosses the clipping
        plane x = w. Let v0=(x0, y0, w0) and let v1=(x1, y1, w1). Then the above
        parametric equation becomes the three component equations
                 x(t) = (1-t) * x0 + t * x1,
                 y(t) = (1-t) * y0 + t * y1,
                 w(t) = (1-t) * w0 + t * w1,  with  0 <= t <= 1.
        Since the clipping plane in this example is x = w, we need to solve the
        equation x(t) = w(t) for t. So we need to solve
                   (1-t) * x0 + t * x1 = (1-t) * w0 + t * w1
        for t. Here are a few algebra steps.
                  x0 - t * x0 + t * x1 = w0 - t * w0 + t * w1
                    x0 + t * (x1 - x0) = w0 + t * (w1 - w0)
         t * (x1 - x0) - t * (w1 - w0) = w0 - x0
           t * [(x1 - x0) - (w1 - w0)] = w0 - x0
                                     t = (w0 - x0)/[(x1 - x0) - (w1 - w0)]
                                       = (x0 - w0)/[(x0 - x1) - (w0 - w1)]
                                       = (x0 - w0)/[(x0 - w0) - (x1 - w1)]
        We get similar equations for t if we clip against the other clipping planes
        (x = -w, y = w, y = -w, or z = w) and we assume that v0 is on the "wrong side"
        and v1 is on the "right side". With the above value for t, the new point p(t)
        that replaces v0 can easily be computed.

                             x=1
                              |
                        v1    |
                          \   |
                            \ |
                              v0=p( (x0-w0)/[(x0-w0) - (x1-w1)] )
                              |
                              |
                              |

         Finally, the new line segment between v1 and the new v0 is recursively
         clipped so that it can be checked once again to see if it can be trivially
         accepted, trivially rejected, or clipped again.
*/
public class Clip
{
   /**
      If the line segment sticks out of the normalized view rectangle,
      then clip it so that it is contained in the normalized view rectangle.
   */
   public static boolean clip(LineSegment ls)
   {
      // make local copies of several values
      Vertex v0 = ls.v[0],  v1 = ls.v[1];
      double x0 = v0.x,     x1 = v1.x;
      double y0 = v0.y,     y1 = v1.y;
      double z0 = v0.z,     z1 = v1.z;
      double w0 = v0.w,     w1 = v1.w;

      double abs_w0 = Math.abs( w0 );
      double abs_w1 = Math.abs( w1 );

      // 1. Check for trivial accept.
      if ( ! ( Math.abs( x0 ) > abs_w0
            || Math.abs( y0 ) > abs_w0
            || Math.abs( x1 ) > abs_w1
            || Math.abs( y1 ) > abs_w1
//          ||             z0 > w0
//          ||             z1 > w1 ) )
            || Math.abs( z0 ) > abs_w0     // this allows objects to reappear after
            || Math.abs( z1 ) > abs_w1 ) ) // they go far enough behind the camera
      {
         //System.err.println("Trivial accept.");
         return true;
      }
      // 2. Check for trivial delete.
      else if ( (w0 - x0 < 0  &&  w1 - x1 < 0) // right of the plane x=w
             || (w0 + x0 < 0  &&  w1 + x1 < 0) // left of the plane x=-w
             || (w0 - y0 < 0  &&  w1 - y1 < 0) // above the plane y=w
             || (w0 + y0 < 0  &&  w1 + y1 < 0) // below the plane y=-w
             || (w0 - z0 < 0  &&  w1 - z1 < 0))// right of the plne z=w
      {
         //System.err.println("Trivial delete.");
         return false;
      }
      // 3. Need to clip this line segment.
      if (w0 - z0 < 0 || w1 - z1 < 0)  // ls crosses the plane z = w
      {
         if (w0 - z0 < 0)  // clip off v0
         {
            // create a new Vertex between v[1] and v[0]
            Vertex v_new = interpolateNewVertex(v1, v0, 5);
            // modify the LineSegment to contain v[1] and the new Vertex
            ls.v[0] = v_new;
         }
         else // (w1 - z1 < 0)  // clip off v1
         {
            // create a new Vertex between v[0] and v[1]
            Vertex v_new = interpolateNewVertex(v0, v1, 5);
            // modify the LineSegment to contain v[0] and the new Vertex
            ls.v[1] = v_new;
         }
      }
      else if (w0 - x0 < 0 || w1 - x1 < 0)  // ls crosses the plane x = w
      {
         if (w0 - x0 < 0)  // clip off v0
         {
            // create a new Vertex between v[1] and v[0]
            Vertex v_new = interpolateNewVertex(v1, v0, 1);
            // modify the LineSegment to contain v[1] and the new Vertex
            ls.v[0] = v_new;
         }
         else // (w1 - x1 < 0)  // clip off v1
         {
            // create a new Vertex between v[0] and v[1]
            Vertex v_new = interpolateNewVertex(v0, v1, 1);
            // modify the LineSegment to contain v[0] and the new Vertex
            ls.v[1] = v_new;
         }
      }
      else if (w0 + x0 < 0 || w1 + x1 < 0)  // ls crosses the plane x = -w
      {
         if (w0 + x0 < 0)  // clip off v0
         {
            // create a new Vertex between v1 and v0
            Vertex v_new = interpolateNewVertex(v1, v0, 2);
            // modify the LineSegment to contain v1 and the new Vertex
            ls.v[0] = v_new;
         }
         else // (w1 + x1 < 0)  // clip off v1
         {
            // create a new Vertex between v0 and v1
            Vertex v_new = interpolateNewVertex(v0, v1, 2);
            // modify the LineSegment to contain v0 and the new Vertex
            ls.v[1] = v_new;
         }
      }
      else if (w0 - y0 < 0 || w1 - y1 < 0)  // ls crosses the plane y = w
      {
         if (w0 - y0 < 0)  // clip off v0
         {
            // create a new Vertex between v1 and v0
            Vertex v_new = interpolateNewVertex(v1, v0, 3);
            // modify the LineSegment to contain v1 and the new Vertex
            ls.v[0] = v_new;
         }
         else // (w1 - y1 < 0)  // clip off v1
         {
            // create a new Vertex between v0 and v1
            Vertex v_new = interpolateNewVertex(v0, v1, 3);
            // modify the LineSegment to contain v0 and the new Vertex
            ls.v[1] = v_new;
         }
      }
      else if (w0 + y0 < 0 || w1 + y1 < 0)  // ls crosses the plane y = -w
      {
         if (w0 + y0 < 0)  // clip off v0
         {
            // create a new Vertex between v1 and v0
            Vertex v_new = interpolateNewVertex(v1, v0, 4);
            // modify the LineSegment to contain v1 and the new Vertex
            ls.v[0] = v_new;
         }
         else // (w1 + y1 < 0)  // clip off v1
         {
            // create a new Vertex between v0 and v1
            Vertex v_new = interpolateNewVertex(v0, v1, 4);
            // modify the LineSegment to contain v0 and the new Vertex
            ls.v[1] = v_new;
         }
      }
      else
      {  // should never get here
         System.err.println("Clipping error!");
         System.exit(-1);
      }
      return clip(ls);  // recursively clip this line segment again
   }


   /**
      This method takes in two vertices, one that is on the "right" side
      of a clipping plane and the other that is on the "wrong" side of the
      clipping plane, and an integer which specifies which clipping plane
      to use, where
         eqn_number == 1  means the clipping plane x =  w
         eqn_number == 2  means the clipping plane x = -w
         eqn_number == 3  means the clipping plane y =  w
         eqn_number == 4  means the clipping plane y = -w
         eqn_number == 5  means the clipping plane z =  w

      This method returns the vertex that is the intersection point
      between the given line segment and the given clipping plane.

      This method solves for the value of t for which the parametric
      equation
                  p(t) = (1-t) * v_outside + t * v_inside
      intersects the given clipping plane (notice that the equation
      is parameterized so that we move from the outside vertex towards
      the inside vertex as t increases from 0 to 1). The solved for
      value of t is then plugged into the parametric formula to get
      the coordinates of the intersection point.
   */
   private static Vertex interpolateNewVertex(Vertex v_inside,
                                              Vertex v_outside,
                                              int eqn_number)
   {
      //System.err.println("Create new vertex.");
      // make local copies of several values
      double vix =  v_inside.x; // "i" for "inside"
      double viy =  v_inside.y;
      double viz =  v_inside.z;
      double viw =  v_inside.w;
      double vox = v_outside.x; // "o" for "outside"
      double voy = v_outside.y;
      double voz = v_outside.z;
      double vow = v_outside.w;
      // interpolate between v_outside and v_inside
      double t = 0.0;
      if (1 == eqn_number)             // clip to plane x = w
         t = -(vow - vox)/( (viw - vix) - (vow - vox) );
      else if (2 == eqn_number)        // clip to plane x = -w
         t = -(vow + vox)/( (viw + vix) - (vow + vox) );
      else if (3 == eqn_number)       // clip to plane y = w
         t = -(vow - voy)/( (viw - viy) - (vow - voy) );
      else if (4 == eqn_number)       // clip to plane y = -w
         t = -(vow + voy)/( (viw + viy) - (vow + voy) );
      else if (5 == eqn_number)       // clip to plane z = w
         t = -(vow - voz)/( (viw - viz) - (vow - voz) );

      t += 0.00001;  /*** keep the new vertex off the edge!! ***/

      Vertex v_new = new Vertex();

      // use the value of t to interpolate the coordinates of the new vertex
      v_new.x = (1-t) * vox + t * vix;
      v_new.y = (1-t) * voy + t * viy;
      v_new.z = (1-t) * voz + t * viz;
      v_new.w = (1-t) * vow + t * viw;
      // use the value of t to interpolate the color of the new vertex
      v_new.r = (1-t) * v_outside.r + t * v_inside.r;
      v_new.g = (1-t) * v_outside.g + t * v_inside.g;
      v_new.b = (1-t) * v_outside.b + t * v_inside.b;

      //System.err.println("Clip: eqn_number = " + eqn_number );
      //System.err.printf("Clip: [% .8f % .8f % .8f % .8f]\n", vix, viy, viz, viw);
      //System.err.printf("Clip: [% .8f % .8f % .8f % .8f]\n", vox, voy, voz, vow);
      //System.err.printf("Clip: t = % .8f\n", t);
      //System.err.printf("Clip: <% .8f % .8f % .8f % .8f>\n", v_new.x,v_new.y,v_new.z,v_new.w);

      return v_new;
   }

}
