/*

*/
package pipeline;
import scene.*;

/**
   Transform line segments from the view coordinate system to the
   (normalized) camera coordinate system.

   This stage transforms the view coordinate system so that the
   camera's view volume becomes the standard view volume.

   There are two standard view volumes, one for perspective projection
   and one for orthographic projection.

   The standard perspective view volume is the infinitely long pyramid
   with its apex at the origin and intersecting the plane z = -1 at the
   corners (-1, -1, -1) and (+1, +1, -1).

   The standard orthographic view volume is the infinitely long
   parallelepiped centered on the z-axis and intersecting the xy-plane
   at the corners (-1, -1, 0) and (+1, +1, 0).

   The current view volume (determined by the Scene's Camera object) is
   either the infinitely long pyramid with its apex at the origin and
   intersecting the plane z = -near at the corners (left, bottom, -near)
   and (right, top, -near), or it is the infinitely long parallelepiped
   parallel to the z-axis and intersecting the xy-plane at the corners
   (left, bottom, 0) and (right, top, 0).

   The view coordinate system is relative to the current view volume.

   The camera coordinate system is relative to the standard view volume.

   The matrices that transform the current view volume into the standard
   view volume also transform the view coordinate system into the camera
   coordinate system.
*/
public class Normalize
{
   /**
      For each line segment in the model, transform each vertex,
      using the camera's normalizing matrix, from view coordinates
      to camera coordinates.

      @param ls LineSegment to project into homogeneous space
      @param normalizeMatrix the projection matrix from the Scene's Camera
   */
   public static void normalize(LineSegment ls, Matrix normalizeMatrix)
   {
      // Mutate each Vertex object so that it contains
      // the camera coordinates.
      ls.v[0].times(normalizeMatrix);
      ls.v[1].times(normalizeMatrix);
   }
}
