/*

*/
package pipeline;
import scene.*;

/**
   Transform line segments from the view coordinate system to the
   (normalized) camera coordinate system.

   This stage transforms the view coordinate system so that the
   camera's view volume becomes the standard view volume.

   There are two standard view volumes, one for perspective projection
   and one for orthographic projection.

   The standard perspective view volume is the infinitely long pyramid
   with its apex at the origin and intersecting the plane z = -1 at the
   corners (-1, -1, -1) and (+1, +1, -1).

   The standard orthographic view volume is the infinitely long
   parallelepiped centered on the z-axis and intersecting the xy-plane
   at the corners (-1, -1, 0) and (+1, +1, 0).

   The current view volume (determined by the Scene's Camera object) is
   either the infinitely long pyramid with its apex at the origin and
   intersecting the plane z = -near at the corners (left, bottom, -near)
   and (right, top, -near), or it is the infinitely long parallelepiped
   parallel to the z-axis and intersecting the xy-plane at the corners
   (left, bottom, 0) and (right, top, 0).

   The view coordinate system is relative to the current view volume.

   The camera coordinate system is relative to the standard view volume.

   The formulas that transform the current view volume into the standard
   view volume also transform the view coordinate system into the camera
   coordinate system.
*/
public class Normalize_formulas
{
   /**
      For each line segment in the model, transform each vertex
      from view coordinates to camera coordinates.
   */
   public static void normalize(LineSegment ls, Camera camera)
   {
      double l = camera.left;   // These numbers define the camera's "view volume".
      double r = camera.right;  // (In future renderers, these numbers will be
      double b = camera.bottom; //  encoded into a "projection matrix" in the
      double t = camera.top;    //  Camera object.)
      double near = - camera.n;

      double v0_x, v0_y, v0_z, v1_x, v1_y, v1_z;

      if ( camera.perspective )
      {
         // We use two steps to transform the camera's perspective view volume
         // into the standard perspective view volume. The first step "skews" the
         // current view volume so that its center line is the negative z-axis.
         // The second step scales the skewed view volume so that it intersects
         // the plane z = -1 with corners (-1, -1, -1) and (+1, +1, -1).

         // For each vertex in the line segment, skew its coordinates so
         // that the negative z-axis is at the center of the view volume.
         v0_z = ls.v[0].z;
         v0_y = ls.v[0].y - v0_z * (t + b)/(2 * near);
         v0_x = ls.v[0].x - v0_z * (r + l)/(2 * near);

         v1_z = ls.v[1].z;
         v1_y = ls.v[1].y - v1_z * (t + b)/(2 * near);
         v1_x = ls.v[1].x - v1_z * (r + l)/(2 * near);

         // For each vertex in the line segment, scale its coordinates so
         // that the view volume has corners (-1, -1, -1) and (+1, +1, -1).
         v0_x = (2 * near * v0_x)/(r - l);
         v0_y = (2 * near * v0_y)/(t - b);

         v1_x = (2 * near * v1_x)/(r - l);
         v1_y = (2 * near * v1_y)/(t - b);
      }
      else
      {
         // We use two steps to transform the camera's orthographic view volume
         // into the standard orthographic view volume. The first step translates the
         // current view volume so that its center line is the z-axis. The second
         // step scales the translated view volume so that it intersects the plane
         // z = 0 with corners (-1, -1, 0) and (+1, +1, 0).

         // For each vertex in the line segment, translate its coordinates
         // so that the z-axis is at the center of the view volume.
         v0_x = ls.v[0].x - (r + l)/2;
         v0_y = ls.v[0].y - (t + b)/2;
         v0_z = ls.v[0].z;

         v1_x = ls.v[1].x - (r + l)/2;
         v1_y = ls.v[1].y - (t + b)/2;
         v1_z = ls.v[1].z;

         // For each vertex in the line segment, scale its coordinates so
         // that the view volume has corners (-1, -1, 0) and (+1, +1, 0).
         v0_x = (2 * v0_x)/(r - l);
         v0_y = (2 * v0_y)/(t - b);

         v1_x = (2 * v1_x)/(r - l);
         v1_y = (2 * v1_y)/(t - b);
      }

      // Mutate the LineSegment object so that it refers to new
      // Vertex objects containing the normalized view coordinates.
      ls.v[0] = new Vertex( v0_x, v0_y, v0_z, ls.v[0].getColor() );
      ls.v[1] = new Vertex( v1_x, v1_y, v1_z, ls.v[1].getColor() );
   }
}
