/*

*/
package scene;

/**
<pre>
   We use two steps to transform the camera's orthographic view volume
   into the standard orthographic view volume. The first step translates
   the  current view volume so that its center line is the z-axis. The
   second step scales the translated view volume so that it intersects
   the plane z = 0 with corners (-1, -1, 0) and (+1, +1, 0).

   This matrix translates the camera's view volume so that the translated
   view volume will be centered on the z-axis.

     [ 1  0  0  -(r+l)/2 ]
     [ 0  1  0  -(t+b)/2 ]
     [ 0  0  1      0    ]
     [ 0  0  0      1    ]

   This matrix scales the translated view volume so that it will
   be 2 units wide and 2 units tall at the view plane z = 0.

     [ 2/(r-l)    0    0  0 ]
     [   0     2/(t-b) 0  0 ]
     [   0        0    1  0 ]
     [   0        0    0  1 ]

   The matrix product looks like this.

     [ 1  0  0  -(r+l)/2 ]   [ 2/(r-l)    0    0  0 ]
     [ 0  1  0  -(t+b)/2 ] = [   0     2/(t-b) 0  0 ]
     [ 0  0  1      0    ]   [   0        0    1  0 ]
     [ 0  0  0      1    ]   [   0        0    0  1 ]

          [ 2/(r-l)    0    0  -(r+l)/2 ]
        = [   0     2/(t-b) 0  -(t+b)/2 ]
          [   0        0    1      0    ]
          [   0        0    0      1    ]

   This product matrix transforms the camera's orthographic
   view volume into the standard (normalized) orthographic
   view volume whose intersection with the view plane, z = 0,
   has left = -1, right = +1, bottom = -1, and top = +1.
</pre>
*/
public class OrthographicNormalizeMatrix extends Matrix
{
   /**
      Construct the matrix that transforms from the camera's
      orthographic view coordinate system to the normalized
      orthographic camera coordinate system.
   */
   public OrthographicNormalizeMatrix(double l, double r, double b, double t)
   {
      super();

      Matrix m1, m2, m3;

      m1 = new Matrix(
               new Vector(  1.0,      0.0,    0.0, 0.0),
               new Vector(  0.0,      1.0,    0.0, 0.0),
               new Vector(  0.0,      0.0,    1.0, 0.0),
               new Vector(-(r+l)/2, -(t+b)/2, 0.0, 1.0));

      m2 = new Matrix(
               new Vector(2/(r-l),   0.0,   0.0, 0.0),
               new Vector(   0.0,  2/(t-b), 0.0, 0.0),
               new Vector(   0.0,    0.0,   1.0, 0.0),
               new Vector(   0.0,    0.0,   0.0, 1.0));

      m3 = m2.times( m1 );

      this.v1 = m3.v1;
      this.v2 = m3.v2;
      this.v3 = m3.v3;
      this.v4 = m3.v4;
   }
}
