/*

*/
package scene;

/**
<pre>
   This is a (non-singular) perspective projection matrix.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  1  1 ]
     [ 0  0 -1  0 ]
   Notice that this matrix maps
        (x, y, z, 1) -> (x, y, z+1, -z)
   so it "projects" the camera space w = 1 onto the
   homogeneous space w + z = 1.

   After the perspective division stage we get
        (x, y, z+1, -z) -> (-x/z, -y/z, -1-(1/z), 1)
   Notice that the standard view plane z = -1 will be sent to z = 0.

   This matrix transforms the standard view pyramid in camera
   coordinates into a homogeneous view pyramid in clip coordinates
   that, after perspective division, will be the standard infinite
   parallelepiped parallel to the z-axis in normalized device
   coordinates (the same infinite parallelepiped used by the
   orthographic projection).

   Consider a projective ray through the origin in camera space,
       (m*t, n*t, -t, 1)  with 0 <= t < +infinity
   When t=1 this ray intersects the standard view plane at the
   point (m, n, -1, 1). This ray gets transformed by the matrix and
   perspective division to
       (m*t, n*t, -t, 1) -> (m*t, n*t, -t+1, t)
                         -> (m, n, -1 + (1/t), 1)  with 0 <= t < +infinity.
   This transformed ray is (part of) a horizontal line in NDC that
   intersects the view plane z = 0 at (m, n, 0, 1) when t=1.

   So the combination of this matrix transformation and perspective
   division sends projective rays from camera space to horizontal lines
   in normalized device coordinates. So the standard infinite view
   pyramid in camera coordinates gets sent to the standard infinite view
   parallelepiped in NDC. Also, the standard view rectangle in camera
   coordinates is sent to the standard view rectangle in NDC.

   Since a solid, three-dimensional, view pyramid is sent to a solid,
   three-dimensional, view parallelepiped, the projection transformation
   based on this matrix preserves depth information, which is what we are
   after. (Our previous perspective projection transformation, based on
   our previous, singular, perspective projection matrix, sent a solid,
   three-dimensional view pyramid to a flat, two-dimensional view rectangle,
   and lost all the depth information.)


   After this projection transformation, we want to skew the homogeneous
   space along the w-axis and in the z-direction so that the near plane,
   z = -near in camera coordinates, is transformed to the homogeneous
   clipping plane w - z = 0. A skew transformation of homogeneous space
   along the w-axis and in the z-direction looks like this.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  1  b ]
     [ 0  0  0  1 ]
   for some b that we need to compute. The projection transformation
   followed by the skew transformation looks like this.
     [ 1  0  0  0 ]   [ 1  0  0  0 ]   [ 1  0  0  0 ]
     [ 0  1  0  0 ] * [ 0  1  0  0 ] = [ 0  1  0  0 ]
     [ 0  0  1  b ]   [ 0  0  1  1 ]   [ 0  0 1-b 1 ]
     [ 0  0  0  1 ]   [ 0  0 -1  0 ]   [ 0  0  -1 0 ]
   Apply the combined projection/skew matrix to the near plane,
   z = -near, from camera coordinates.
     [ 1  0  0  0 ]   [   x   ]   [     x       ]
     [ 0  1  0  0 ] * [   y   ] = [     y       ]
     [ 0  0 1-b 1 ]   [ -near ]   [-near*(1-b)+1]
     [ 0  0  -1 0 ]   [   1   ]   [    near     ]
   After perspective division, this becomes
     ( x/near, y/near, -(1-b)+(1/near), 1).
   We want this to be the z = 1 clipping plane in NDC.
   So we need to solve
       -(1-b) + 1/near = 1
       -1 + b + 1/near = 1
            b + 1/near = 2
            b = 2 - 1/near
              = (2*near - 1)/near
</pre>
*/
public class PerspectiveProjectionMatrix0 extends Matrix
{
   /**
   <pre>
      Construct this non-singular perspective projection matrix.
        [ 1  0  0  0 ]
        [ 0  1  0  0 ]
        [ 0  0  b  1 ]
        [ 0  0 -1  0 ]
      where
        b = -1 + (1/near)
   </pre>
   */
   public PerspectiveProjectionMatrix0(double n)
   {
      super(  new Vector(1.0,  0.0,  0.0,      0.0),
              new Vector(0.0,  1.0,  0.0,      0.0),
              new Vector(0.0,  0.0, -1+(1/n), -1.0),
              new Vector(0.0,  0.0,  1.0,      0.0) );
   }
}
