/*

*/
package scene;

/**
<pre>
   This is a (non-singular) perspective projection matrix.
     [ 1  0  0  0 ]
     [ 0  1  0  0 ]
     [ 0  0  0  1 ]
     [ 0  0 -1  0 ]
   Notice that this matrix maps
        (x, y, z, 1) -> (x, y, 1, -z)
   so it "projects" the three-dimensional camera space w = 1
   onto the three-dimensional homogeneous space z = 1 in clip
   coordinates.

   We want our projection matrix to transform the near plane,
   z = -near in camera coordinates, to the clipping plane z = +1
   in NDC. Notice that
     [ 1  0  0   0   ]
     [ 0  1  0   0   ]
     [ 0  0  0  near ]
     [ 0  0 -1   0   ]
   maps
        (x, y, z, 1) -> (x, y, near, -z)
   so it "projects" the three-dimensional camera space w = 1
   onto the three-dimensional homogeneous space z = near in
   clip coordinates.

   Notice that the camera's near plane, z = -near, gets transformed
   as
        (x, y, -near, 1) -> (x, y, near, near)
   so the transformed near plane satisfies the equation w - z = 0
   in clip coordinates.

   After the perspective division stage we get
        (x, y, z, 1) -> (x, y, near, -z) -> (-x/z, -y/z, -near/z, 1)
   Notice that the camera's near plane, z = -near, will be
   transformed to z = +1 in NDC.

   Consider a projective ray through the origin in camera space,
       (m*t, n*t, -t, 1)  with 0 <= t < +infinity.
   When t=near this ray intersects the camera's near plane at the
   point (m*near, n*near, -near, 1). This ray gets transformed by
   this matrix and perspective division to
       (m*t, n*t, -t, 1) -> (m*t, n*t, near, t)
                         -> (m, n, near/t, 1).
   When t=near the transformed ray intersects the plane z = +1 in NDC.
   When t < near, near/t > 1, so the part of the ray in front of the near
   plane (i.e., t < near) will satisfy the inequality z > w in clip
   coordinates, or w - z < 0. Similarly, the part of the ray behind the
   near plane (i.e., t > near) will satisfy the inequality w - z > 0 in
   clip coordinates.

   So for the clipping stage to clip off the part of a scene that is in
   front of the camera's near plane, the clipping algorithm should keep
   any vertex in clip coordinates that satisfies the inequality w - z >= 0,
   and discard any vertex that satisfies the inequality w - z < 0.
</pre>
*/
public class PerspectiveProjectionMatrix1 extends Matrix
{
   /**
   <pre>
      Construct this non-singular perspective projection matrix.
        [ 1  0  0  0  ]
        [ 0  1  0  0  ]
        [ 0  0  0 near]
        [ 0  0 -1  0  ]
   </pre>
   */
   public PerspectiveProjectionMatrix1(double near)
   {
      super(  new Vector(1.0,  0.0,  0.0,  0.0),
              new Vector(0.0,  1.0,  0.0,  0.0),
              new Vector(0.0,  0.0,  0.0, -1.0),
              new Vector(0.0,  0.0, near,  0.0) );
   }
}
